package it.eng.spago.cms.access.verifiers;

import it.eng.spago.base.Constants;
import it.eng.spago.base.SourceBean;
import it.eng.spago.cms.util.Path;
import it.eng.spago.configuration.ConfigSingleton;
import it.eng.spago.tracing.TracerSingleton;
import java.util.List;
import java.util.Vector;

/**
 *
 * Implements methods for verify permission of the user 
 * to execute operation on the repository.
 * JSR 170 specification define four operation which can be 
 * under permission control:
 * <ul>
 * <li>Add nodes</li>
 * <li>Set property</li>
 * <li>Erase nodes</li>
 * <li>Read Nodes</li>
 * </ul>
 * The class use the configuration file "contentPermission.xml" 
 * and the Spago IEenUserProfile interface to control the permission 
 * of the user. Each rule of the user 
 *(See Spago IEngUserProfile) is mapped to a set of the four repository
 * operation on different path. 
 * 
 */
public class XMLPermissionVerifier {
     
	SourceBean permissionsConf  = null;

	public void init(String nameRep) {
		Object result = ConfigSingleton.getInstance().getFilteredSourceBeanAttribute("CONTENTPERMISSIONS.CONTENTREPOSITORY", "name", nameRep);
	    if(result == null) {
	    	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
            "XMLPermissionVerifier::init:configuration for repository "+nameRep+" not found");
	    }
	    else if(result instanceof Vector) {
	    	TracerSingleton.log(Constants.NOME_MODULO, TracerSingleton.WARNING,
	        "XMLPermissionVerifier::init:exists more than one configuration for the repository "+nameRep);
	    }
	    else {
	    	permissionsConf = (SourceBean)result;
	    }
	}
	
	
	/**
	 * Verify if a user can add a node to a path. The method use
	 * the set of functionalities of the user to verify if at least
	 * one of them permits to add the node.
	 *  
	 * @param path, path of the node father
	 * @param functs, functionalities set of the user 	
	 * @return boolean, if true the user can perform the operation 
	 * false otherwise 
	 */
	public boolean canAddNode(String path, String[] roles) {
		return controlPermission(path, roles, "ADDNODEPERMISSION");
	}
	
	/**
	 * Verify if a user can set properties to a path. The method use
	 * the set of functionalities of the user to verify if at least
	 * one of them permits to set properties.
	 *  
	 * @param path, path of the node to add properties
	 * @param functs, functionalities set of the user 	
	 * @return boolean, if true the user can perform the operation 
	 * false otherwise 
	 */
	public boolean canSetProperty(String path, String[] roles) {
		return controlPermission(path, roles, "SETPROPERTYPERMISSION");
	}
	
	/**
	 * Verify if a user can remove a node. The method use
	 * the set of functionalities of the user to verify if at least
	 * one of them permits to remove the node.
	 *  
	 * @param path, path of the node to remove
	 * @param functs, functionalities set of the user 	
	 * @return boolean, if true the user can perform the operation 
	 * false otherwise 
	 */
	public boolean canRemove(String path, String[] roles) {
		return controlPermission(path, roles, "REMOVEITEMPERMISSION");
	}
	
	/**
	 * Verify if a user can read information of a node. The method use
	 * the set of functionalities of the user to verify if at least
	 * one of them permits to read the node.
	 *  
	 * @param path, path of the node to read
	 * @param functs, functionalities set of the user 	
	 * @return boolean, if true the user can perform the operation 
	 * false otherwise 
	 */
	public boolean canRead(String path, String[] roles) {
		return controlPermission(path, roles, "READITEMPERMISSION");
	}
	
	/**
	 * Verifiy if a user can perform one of the four repository
	 * operation on a specific path. 
	 * 
	 * @param pathStr, node path of the operation  
	 * @param functs, functionalities set of the user
	 * @param operation, string that identify one of the four operation
	 * @return boolean, if true the user can perform the operation 
	 * false otherwise 
	 */
	private boolean controlPermission(String pathStr, String[] roles, String operation) {
    	    
		boolean can = false;
		List confForFunctList = null;
		SourceBean confForFunct = null;
		SourceBean confOp = null;
		List permissionPathList = null;
		SourceBean permissionPathSB = null;
		String permissionPathStr = null;
		Path permissionPath = null;
		String subtreeStr =  null;
		boolean subtree = false;
		Path path = null;
	    
		if(permissionsConf == null) {
			return false;
		}
		
		// create pathobject form path string in input
		try {
			path = Path.create(pathStr);
		} catch(Exception e) { return false; }
		// for all the functionalities control if user have permission
		for(int i=0; i<roles.length; i++) {
			String funct = roles[i];
			// get the sourceBean relative to functionality
			confForFunctList = permissionsConf.getFilteredSourceBeanAttributeAsList("FUNCTIONALITY", "name", funct);
			// if there's no configuration for the functionality or there's more than one conf continue with the next
			if(confForFunctList.size() != 1) {
				continue;
			}
			// get the only configuration sourceBean
			confForFunct = (SourceBean)confForFunctList.get(0);
			// get the sourceBean for the operation request
			confOp = (SourceBean)confForFunct.getAttribute(operation);
			// get under vector sourceBean for path element
			permissionPathList = confOp.getAttributeAsList("PATH");
			// for all path element make control 
			for(int j=0; j<permissionPathList.size(); j++) {
				// get sourceBean of path
				permissionPathSB = (SourceBean)permissionPathList.get(j);
				// get the string path
				permissionPathStr = (String)permissionPathSB.getAttribute("value");
				// if path is malformed continue with the next path
				if(Path.isMalformedPathString(permissionPathStr)) {
					continue;
				}
				// instantiate Path object 
				try {
					permissionPath = Path.create(permissionPathStr);
				} catch (Exception e) { continue; }
				// retrive subtree flag
				subtreeStr = (String)permissionPathSB.getAttribute("subtree");
				if((subtreeStr != null) && (subtreeStr.equalsIgnoreCase("true")) ) {
					subtree= true;
				} else {
					subtree = false;
				}
				// control path
				if(path.equals(permissionPath)) {
					can = true;
					return can;
				}
				if(subtree) {
					if(permissionPath.isAncestorOf(path)) {
						can = true;
						return can;
					}
				}
			}				
		}
        return can;
	}
}
