/*
 * Copyright 2004-2005 The Apache Software Foundation or its licensors,
 *                     as applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.jackrabbit.core.nodetype;

import it.eng.javax.jcr.RepositoryException;
import it.eng.javax.jcr.Value;
import it.eng.javax.jcr.nodetype.PropertyDefinition;

import org.apache.jackrabbit.core.value.InternalValue;
import org.apache.jackrabbit.core.NamespaceResolver;
import org.apache.log4j.Logger;


/**
 * This class implements the <code>PropertyDefinition</code> interface.
 * All method calls are delegated to the wrapped {@link PropDef},
 * performing the translation from <code>QName</code>s to JCR names
 * (and vice versa) where necessary.
 */
public class PropertyDefinitionImpl extends ItemDefinitionImpl
        implements PropertyDefinition {

    /**
     * Logger instance for this class
     */
    private static Logger log = Logger.getLogger(PropertyDefinitionImpl.class);

    /**
     * Package private constructor
     *
     * @param propDef    property definition
     * @param ntMgr      node type manager
     * @param nsResolver namespace resolver
     */
    PropertyDefinitionImpl(PropDef propDef, NodeTypeManagerImpl ntMgr,
                           NamespaceResolver nsResolver) {
        super(propDef, ntMgr, nsResolver);
    }

    /**
     * Returns the wrapped property definition.
     *
     * @return the wrapped property definition.
     */
    public PropDef unwrap() {
        return (PropDef) itemDef;
    }

    //---------------------------------------------------< PropertyDefinition >
    /**
     * {@inheritDoc}
     */
    public Value[] getDefaultValues() {
        InternalValue[] defVals = ((PropDef) itemDef).getDefaultValues();
        if (defVals == null) {
            return null;
        }
        Value[] values = new Value[defVals.length];
        for (int i = 0; i < defVals.length; i++) {
            try {
                values[i] = defVals[i].toJCRValue(nsResolver);
            } catch (RepositoryException re) {
                // should never get here
                String propName = (getName() == null) ? "[null]" : getName();
                log.error("illegal default value specified for property "
                        + propName + " in node type " + getDeclaringNodeType(),
                        re);
                return null;
            }
        }
        return values;
    }

    /**
     * {@inheritDoc}
     */
    public int getRequiredType() {
        return ((PropDef) itemDef).getRequiredType();
    }

    /**
     * {@inheritDoc}
     */
    public String[] getValueConstraints() {
        ValueConstraint[] constraints = ((PropDef) itemDef).getValueConstraints();
        if (constraints == null || constraints.length == 0) {
            return new String[0];
        }
        String[] vca = new String[constraints.length];
        for (int i = 0; i < constraints.length; i++) {
            vca[i] = constraints[i].getDefinition(nsResolver);
        }
        return vca;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isMultiple() {
        return ((PropDef) itemDef).isMultiple();
    }
}

